from TB4hooky.netLoader.MetaFinder import RemoteMetaFinder
from TB4hooky.netLoader.SourceCodeManager import SourceCodeManager
import sys

_meta = {}
_scm = SourceCodeManager()


def register(endpoint: str) -> None:
    if not hasattr(_meta, endpoint):
        finder = RemoteMetaFinder(endpoint, _scm)
        _meta[endpoint] = finder
        sys.meta_path.append(finder)


def unload(endpoint: str) -> None:
    if hasattr(_meta, endpoint):
        finder = _meta.pop(endpoint)
        sys.meta_path.remove(finder)



import inspect


class Inspect(object):
    """Inspect function type"""
    @staticmethod
    def check_function_type(_obj):
        if not inspect.isfunction(_obj) and not inspect.ismethod(_obj):
            raise TypeError(f"Error type is {type(_obj)} should be 'function'.")

    @staticmethod
    def check_string_type(_obj):
        if not isinstance(_obj, str):
            raise TypeError(f"Error type is {type(_obj)} should be 'str' because you use the remote hook.")




from importlib.abc import (
    SourceLoader,
)
import sys
from types import ModuleType
from TB4hooky.netLoader.SourceCodeManager import SourceCodeManager


class RemoteModuleLoader(SourceLoader):
    def __init__(self, endpoint: str, source_code_manager: SourceCodeManager)-> None:
        self._endpoint = endpoint
        self._source_code_manager = source_code_manager

    def load_module(self, fullname: str) -> ModuleType:
        code = self.get_code(fullname)
        mod = sys.modules.setdefault(fullname, ModuleType(fullname))
        mod.__file__ = self.get_filename(fullname)
        mod.__loader__ = self
        mod.__package__ = fullname.rpartition('.')[0]
        exec(code, mod.__dict__)
        return mod

    def get_code(self, fullname: str) -> str:
        filename = self.get_filename(fullname)
        return self._source_code_manager.get_complied_code(filename)

    def get_data(self, path: str) -> bytes: ...

    def get_filename(self, fullname: str) -> str:
        basename = fullname.split('.')[-1]
        return f"{self._endpoint}/{basename}.py"

    def get_source(self, fullname: str) -> str:
        filename = self.get_filename(fullname)
        try:
            return self._source_code_manager.get_source_code(fullname, filename)
        except Exception as _:
            raise ImportError(f"Can't import {filename}.")

    def is_package(self, fullname: str) -> bool:
        return False



from TB4hooky.netLoader.ModuleLoader import RemoteModuleLoader
from types import ModuleType


class RemotePackageLoader(RemoteModuleLoader):
    def load_module(self, fullname: str) -> ModuleType:
        mod = super(RemoteModuleLoader, self).load_module(fullname)
        mod.__path__ = [self._endpoint]
        mod.__package__ = fullname

        return mod

    def get_filename(self, fullname: str) -> str:
        return f"{self._endpoint}/__init__.py"

    def is_package(self, fullname: str) -> bool:
        return True



import base64
import json
import sys
import types

LOGGER = True

from TB4hooky.netLoader.AdapterConnectionRemote import AdapterConnectionRemote as ConnectionRemote
from abc import ABCMeta, abstractmethod
from TB4hooky.typeInspect.Inspect import Inspect
from loguru import logger

DEFAULT_CACHE_COUNT = 64


def unable_logger():
    global LOGGER
    LOGGER = False


class MetaCodeHooker(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def extract_co_code(*args, **kwargs): ...

    @abstractmethod
    def swap_code_info(self, *args, **kwargs): ...

    @abstractmethod
    def swap_remote_code_info(self, *args, **kwargs): ...


class BaseCodeHooker(object):
    """ BaseCodeHooker """

    def __init__(self, _f, remote: bool = False, *args, **kwargs):
        if not ((3, 7) < sys.version_info < (3, 8)):
            raise NotImplemented("Not implemented yet")
        # 远程hook
        self._remote = remote
        # 实现函数对象
        self._f = _f
        # 被hook对象
        self._hook_obj = None
        # 缓存
        self._cache = None
        # 缓存计数器
        self._cache_count = None
        # 当前被hook的函数
        # remote func argument
        self.arg = args
        self.kwarg = kwargs

    # extract implement hook function 'co_code' byte code
    @staticmethod
    def extract_co_code(func):
        return BaseCodeHooker._extract_co_code(func)

    @staticmethod
    def _extract_co_code(func):
        Inspect.check_function_type(func)
        return func.__code__.co_code

    def _swap_code_info(self):
        Inspect.check_function_type(self._f)
        Inspect.check_function_type(self._hook_obj)
        _f_code = self._f.__code__
        if not self._cache:
            self._cache = BaseCodeHooker.extract_co_code(self._f)
        self._hook_obj.__code__ = types.CodeType(
            _f_code.co_argcount,
            _f_code.co_kwonlyargcount,
            _f_code.co_nlocals,
            _f_code.co_stacksize,
            _f_code.co_flags,
            self._cache,
            _f_code.co_consts,
            _f_code.co_names,
            _f_code.co_varnames,
            _f_code.co_filename,
            _f_code.co_name,
            _f_code.co_firstlineno,
            _f_code.co_lnotab,
            _f_code.co_freevars
        )

    def swap_code_info(self):
        self._swap_code_info()

    def swap_remote_code_info(self, host):
        pass


class CodeHooker(BaseCodeHooker):
    def __init__(self, _f, remote: bool = None):
        if remote:
            self._cache_count = DEFAULT_CACHE_COUNT
        super().__init__(_f, remote)

    def __call__(self, func):
        self._hook_obj = func

        def arg_recv(*args, **kwargs):
            if not self._remote:
                if LOGGER:
                    logger.success(f"Hook mod [Local], Local function"
                                   f" from [{self._f.__name__}] to Hook object [{self._hook_obj.__name__}]")

                self.swap_code_info()
            else:
                if LOGGER:
                    logger.success(f"Hook mod [Remote], Remote code "
                                   f"from host [{self._f}] to Hook object [{self._hook_obj.__name__}]")

                self.swap_remote_code_info(self._f)
            return self._hook_obj(*args, **kwargs)

        return arg_recv

    @classmethod
    def serialize_init(cls, func):
        return cls(func, remote=False)

    def _serialize_func(self, count: int) -> str:
        _f_struct = {"cache_count": count,
                     "co_argcount": self._f.__code__.co_argcount,
                     'co_kwonlyargcount': self._f.__code__.co_kwonlyargcount,
                     'co_nlocals': self._f.__code__.co_nlocals,
                     'co_stacksize': self._f.__code__.co_stacksize,
                     'co_flags': self._f.__code__.co_flags,
                     'co_codestring': base64.b64encode(self._f.__code__.co_code).decode(),
                     'co_consts': self._f.__code__.co_consts,
                     'co_names': self._f.__code__.co_names,
                     'co_varnames': self._f.__code__.co_varnames,
                     'co_filename': self._f.__code__.co_filename,
                     'co_name': self._f.__code__.co_name,
                     'co_firstlineno': self._f.__code__.co_firstlineno,
                     'co_lnotab': base64.b64encode(self._f.__code__.co_lnotab).decode(),
                     'co_freevars': self._f.__code__.co_freevars}
        return json.dumps(_f_struct)

    def serialize_func(self, count: int) -> str:
        return self._serialize_func(count)

    def _set_cache(self, serialize: dict):
        for key in serialize.keys():
            value = serialize[key]
            if key == "cache_count":
                self._cache_count = int(value)
            if key in ("co_lnotab", "co_codestring"):
                value = base64.b64decode(value.encode())
            if key in ("co_argcount", "co_kwonlyargcount", "co_nlocals",
                       "co_stacksize", "co_flags", "co_firstlineno"):
                value = int(value)
            if key in ("co_consts", "co_names", "co_varnames", "co_freevars"):
                value = tuple(value)
            self._cache[key] = value

    def swap_remote_code_info(self, host, *args, **kwargs):
        self._swap_remote_code_info(host)

    def _swap_remote_code_info(self, host, *args, **kwargs):
        Inspect.check_string_type(host)
        if not self._cache_count:

            # remote cache
            with ConnectionRemote() as conn:
                remote_content = conn.get_remote_f(host, *args, **kwargs)

            # 判断分发服务器是否在线
            if 0 < (remote_content.status_code - 200) < 10:
                raise Exception(f"Remote code request Error status "
                                f"code: {remote_content.status_code} remote host: {host}.")
            remote_content = remote_content.json()

            # 取出RPC 填充 cache
            self._cache = {}
            self._set_cache(remote_content)
            self._cache.pop("cache_count")

        self._cache_count -= 1
        self._hook_obj.__code__ = types.CodeType(
            self._cache['co_argcount'],
            self._cache['co_kwonlyargcount'],
            self._cache['co_nlocals'],
            self._cache['co_stacksize'],
            self._cache['co_flags'],
            self._cache['co_codestring'],
            self._cache['co_consts'],
            self._cache['co_names'],
            self._cache['co_varnames'],
            self._cache['co_filename'],
            self._cache['co_name'],
            self._cache['co_firstlineno'],
            self._cache['co_lnotab'],
            self._cache['co_freevars']
        )


class InstanceCodeHooker(CodeHooker):
    _INSTANCE = {}

    def __init__(self, _f, instance):
        super().__init__(_f, remote=False)
        # 保证注册过的instance
        if instance not in self._INSTANCE.values():
            self._instance = instance
            self._INSTANCE[hash(instance)] = instance
        else:
            self._instance = self._INSTANCE[hash(instance)]
        Inspect.check_string_type(_f)
        self._f = getattr(instance, _f)

    def __call__(self, func):
        self._hook_obj = func

        def arg_recv(*args, **kwargs):
            if logger:
                logger.success(f"Hook mod [Local], Local function"
                               f" from [{self._instance.__class__.__name__}().{self._f.__name__}] to Hook object [{self._hook_obj.__name__}]")

            self.swap_code_info()

            return self._hook_obj(self._instance, *args, **kwargs)

        return arg_recv

    def swap_remote_code_info(self, host, *args, **kwargs):
        raise NotImplemented("Error 'InstanceCodeHooker' "
               "not support remote hook should use 'CodeHooker'.")

    def _swap_remote_code_info(self, host, *args, **kwargs):
        raise NotImplemented("Error 'InstanceCodeHooker' "
                             "not support remote hook should use 'CodeHooker'.")



"""
coding: utf-8
@Software: PyCharm
@Time:  8:08
@Author: Fake77
@Module Name:
"""
from TB4hooky.netLoader.AdapterConnectionRemote import AdapterConnectionRemotePackage as ConnectionRemotePackage


class SourceCodeManager:
    def __init__(self) -> None:
        self._sources = {}
        self._complied_codes = {}

    def get_source_code(self, filename: str) -> str:
        source = self._sources.get(filename)
        if not source:
            with ConnectionRemotePackage() as conn:
                source = conn.get_remote_package(filename).text
                self._sources[filename] = source
        return source

    def get_complied_code(self, filename: str):
        source = self.get_source_code(filename)
        complied_code = self._complied_codes.get(filename)
        if not complied_code:
            complied_code = compile(source, filename, 'exec')
            self._complied_codes[filename] = complied_code
        return complied_code

    def clear(self) -> None:
        self._sources.clear()
        self._complied_codes.clear()



from importlib.abc import MetaPathFinder
from TB4hooky.netLoader.SourceCodeManager import SourceCodeManager
from TB4hooky.Utils.LinkDealer import LinksDealer
from TB4hooky.netLoader.PackageLoader import RemotePackageLoader
from TB4hooky.netLoader.ModuleLoader import RemoteModuleLoader


class RemoteMetaFinder(MetaPathFinder):
    def __init__(self, endpoint: str, source_code_manager: SourceCodeManager) -> None:
        self._endpoint = endpoint
        self._nodes = {}
        self._source_code_manager = source_code_manager

    def _is_module(self, filename: str, endpoint: str) -> bool:
        return filename in self._nodes[endpoint]

    def _is_package(self, basename: str, endpoint) -> bool:
        return basename in self._nodes[endpoint]

    def _store_node(self, node: str) -> None:
        if not hasattr(self._nodes, node):
            self._nodes[node] = LinksDealer.get_link(node)

    def _get_node(self, path: str):
        if path is None:
            return self._endpoint
        else:
            if not path[0].startswith(self._endpoint):
                return None
        return path[0]

    def find_module(self, fullname: str, path: str = None):
        # 1. 获取当前节点
        current_code = self._get_node(path)
        # 如果节点不存在中止流程
        if not current_code:
            return None
        #   1.1 缓存节点
        self._store_node(current_code)
        # 2. 获取当前的basename (模块名称)
        basename = fullname.split(".")[-1]
        # 3. 处理package
        if self._is_package(basename, current_code):
            child_location = f"{current_code}/{basename}"
            loader = RemotePackageLoader(child_location, self._source_code_manager)
            try:
                loader.load_module(fullname)
                self._nodes[child_location] = LinksDealer.get_link(child_location)
            except ImportError as _:
                loader = None
            return loader
        # 4. 处理module
        filename = f"{basename}.py"
        if self._is_module(filename, current_code):
            return RemoteModuleLoader(current_code, self._source_code_manager)
        else:
            return None

    def invalidate_caches(self) -> None:
        self._nodes.clear()
        self._source_code_manager.clear()



"""
coding: utf-8
@Software: PyCharm
@Time:  7:48
@Author: Fake77
@Module Name:
"""
import setuptools

with open("README.MD", 'r', encoding='utf-8') as des:
    long_description = des.read()


setuptools.setup(
    name='TB4hooky',
    version='0.1.7',
    author='Fake77',
    author_email="yankail520@gmail.com",
    description='Python hook framework support Local hook and Remote hook',
    long_description=long_description,
    long_description_content_type="text/markdown; charset=UTF-8;",
    url='https://github.com/fateofdate/TB4hooky',
    include_package_data=False,
    install_requires=[
        "requests==2.31.0",
        "urllib3==1.26.6",
        "loguru==0.7.2",
        "win32-setctime==1.1.0",
        "wincertstore==0.2",
        "charset-normalizer==3.3.2",
        "colorama==0.4.6",
        "idna==3.6",
    ],

    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        'Operating System :: OS Independent'
    ],
    python_requires=">=3.7"
)


from html.parser import HTMLParser
import requests


class LinksDealer(object):
    @staticmethod
    def _get_links(addr):
        links = set()

        class LinksParser(HTMLParser):
            def handle_starttag(self, tag, attrs):
                if tag == 'a':
                    attrs = dict(attrs)
                    links.add(attrs.get('href').rstrip('/'))

        try:
            content = requests.get(addr).text
            parser = LinksParser()
            parser.feed(content)
        except Exception as _:
            pass
        return links

    @staticmethod
    def get_link(addr):
        return LinksDealer._get_links(addr)

    @staticmethod
    def _filter_links(links):
        return links.startswith(("http://", "https://"))

    @staticmethod
    def filter_links(links):
        return LinksDealer._filter_links(links)
